# Copyright (c) Qotto, 2021

from typing import Union

from eventy.record import Record, Event, Request, Response
from eventy.serialization import RecordSerializer


class DummySerializer(RecordSerializer):
    """
    Warning: for tests purposes, does not encode in bytes.
    """

    def encode(self, record: Record) -> bytes:
        return record  # type: ignore

    def decode(self, encoded: bytes) -> Union[Event, Request, Response]:
        return encoded  # type: ignore



# Copyright (c) Qotto, 2021

"""
Request record
"""

from eventy.record import Record

__all__ = [
    'Request',
]


class Request(Record):
    """
    Request implementation of the Record abstract base class
    """

    @property
    def type(self):
        """
        Record type (REQUEST)

        :return: "REQUEST"
        """

        return 'REQUEST'



# Copyright (c) Qotto, 2021

"""
Response record
"""

from datetime import datetime
from typing import Dict, Any, Optional
from uuid import UUID

from eventy.record.errors import RecordAttributeTypeError, RecordAttributeValueError
from eventy.record.record import Record

__all__ = [
    'Response',
]


class Response(Record):
    """
    Response implementation of the Record abstract base class
    """

    def __init__(
        self,
        protocol_version: str,
        namespace: str,
        name: str,
        version: str,
        source: str,
        destination: str,
        request_uuid: str,
        uuid: Optional[str] = None,
        ok: bool = True,
        error_code: Optional[int] = None,
        error_message: Optional[str] = None,
        correlation_id: Optional[str] = None,
        partition_key: Optional[str] = None,
        date: Optional[datetime] = None,
        context: Optional[Dict[str, Dict[str, Any]]] = None,
        data: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        Initialize Record attributes

        :param protocol_version: Version of the Eventy protocol in SemVer format
        :param namespace: Namespace of the record schema in URN format
        :param name: Name of the record schema
        :param version: Version of the record schema in SemVer format
        :param source: Emitter of the record in URN format
        :param destination: Destination of the response, should be source of corresponding request, in URN format
        :param request_uuid: UUID v4 of the corresponding request
        :param uuid: UUID v4 of the response. Default: auto-generated
        :param ok:
        :param error_code:
        :param error_message:
        :param correlation_id: Correlation ID, propagated. Default: auto-generated
        :param partition_key: Partition or database key. Default: None
        :param date: Date of the event. Default: now
        :param context: Context data dictionary, propagated. Default: None
        :param data: Domain data dictionary. Default: None

        :raises RecordAttributeError:
        """
        super().__init__(
            protocol_version,
            namespace,
            name,
            version,
            source,
            uuid,
            correlation_id,
            partition_key,
            date,
            context,
            data,
        )
        self.destination = destination
        self.request_uuid = request_uuid
        self.ok = ok
        self.error_code = error_code
        self.error_message = error_message

    @property
    def type(self):
        """
        Record type (RESPONSE)

        :return: "RESPONSE"
        """
        return 'RESPONSE'

    @property
    def destination(self) -> str:
        return self._destination

    @destination.setter
    def destination(self, destination: str) -> None:
        self._destination = destination

    @property
    def request_uuid(self) -> str:
        return self._request_uuid

    @request_uuid.setter
    def request_uuid(self, request_uuid: str) -> None:
        try:
            UUID(request_uuid)
        except TypeError as e:
            raise RecordAttributeValueError('request_uuid', request_uuid, str(e)) from e
        except ValueError as e:
            raise RecordAttributeValueError('request_uuid', request_uuid, str(e)) from e
        self._request_uuid = request_uuid

    @property
    def ok(self) -> bool:
        return self._ok

    @ok.setter
    def ok(self, ok: bool) -> None:
        if not isinstance(ok, bool):
            raise RecordAttributeTypeError('ok', bool, ok)
        self._ok = ok

    @property
    def error_code(self) -> Optional[int]:
        return self._error_code

    @error_code.setter
    def error_code(self, error_code) -> None:
        # TODO: Check if ok is True? This would impose to update ok before error_code, but why not?
        if error_code is not None and not isinstance(error_code, int):
            raise RecordAttributeTypeError('error_code', 'int or None', error_code)
        self._error_code = error_code

    @property
    def error_message(self) -> Optional[str]:
        return self._error_message

    @error_message.setter
    def error_message(self, error_message) -> None:
        # TODO: Check if ok is True? This would impose to update ok before error_message, but why not?
        if error_message is not None and not isinstance(error_message, str):
            raise RecordAttributeTypeError('error_message', 'str or None', error_message)
        self._error_message = error_message



# Copyright (c) Qotto, 2021

"""
Requests integration utilities

Utility functions to integrate the eventy protocol in apps using requests.

Usage::

    import eventy.integration.requests
    eventy.integration.requests.get('https://httpbin.org/get')
"""

import requests.api as vendor_requests_api
import requests.models as vendor_requests_models
import requests.sessions as vendor_requests_sessions

from eventy.trace_id import correlation_id_var

__all__ = [
    'Session',
    'session',
    'request',
    'get',
    'options',
    'head',
    'post',
    'put',
    'patch',
    'delete',
]


def _merge_docs(eventy_object, vendor_object) -> None:
    eventy_doc: str = getattr(eventy_object, '__doc__', '')
    vendor_doc: str = getattr(vendor_object, '__doc__', '').replace('>>>', '>test>')
    if vendor_doc:
        vendor_doc = f'\nVENDOR DOC:\n\n{vendor_doc}'
    merged_doc = eventy_doc + vendor_doc
    setattr(eventy_object, '__doc__', merged_doc)


class Session(vendor_requests_sessions.Session):
    """
    Modified implementation of vendor requests Session, propagating correlation_id.

    This class override ``prepare_request`` in order to insert x-correlation-id header.
    """

    def prepare_request(
        self,
        request: vendor_requests_models.Request
    ) -> vendor_requests_models.PreparedRequest:
        """
        Insert x-correlation-id header before calling vendor requests method.
        """
        if not request.headers:
            request.headers = {}
        request.headers.update({'x-correlation-id': correlation_id_var.get()})
        return super().prepare_request(request)


def session() -> Session:
    """
    Create a modified version of vendor Session, propagating correlation id.

    This function instantiates and returns a Session instance.
    """
    return Session()


def request(method, url, **kwargs):
    """
    Send a request propagating correlation id.

    Keywords arguments are propagated to vendor implementation.
    """
    # Code is directly from vendor requests, using our modified Session class here
    with Session() as session:
        return session.request(method=method, url=url, **kwargs)


def get(url, params=None, **kwargs):
    """
    Send a GET request propagating correlation id.

    Keywords arguments are propagated to vendor implementation.
    """
    # Code is directly from vendor requests, using our modified request function here
    return request('get', url, params=params, **kwargs)


def options(url, **kwargs):
    """
    Send a OPTIONS request propagating correlation id.

    Keywords arguments are propagated to vendor implementation.
    """
    # Code is directly from vendor requests, using our modified request function here
    return request('options', url, **kwargs)


def head(url, **kwargs):
    """
    Send a HEAD request propagating correlation id.

    Keywords arguments are propagated to vendor implementation.
    """
    # Code is directly from vendor requests, using our modified request function here
    kwargs.setdefault('allow_redirects', False)
    return request('head', url, **kwargs)


def post(url, data=None, json=None, **kwargs):
    """
    Send a POST request propagating correlation id.

    Keywords arguments are propagated to vendor implementation.
    """
    # Code is directly from vendor requests, using our modified request function here
    return request('post', url, data=data, json=json, **kwargs)


def put(url, data=None, **kwargs):
    """
    Send a PUT request propagating correlation id.

    Keywords arguments are propagated to vendor implementation.
    """
    # Code is directly from vendor requests, using our modified request function here
    return request('put', url, data=data, **kwargs)


def patch(url, data=None, **kwargs):
    """
    Send a PATCH request propagating correlation id.

    Keywords arguments are propagated to vendor implementation.
    """
    # Code is directly from vendor requests, using our modified request function here
    return request('patch', url, data=data, **kwargs)


def delete(url, **kwargs):
    """
    Send a DELETE request propagating correlation id.

    Keywords arguments are propagated to vendor implementation.
    """
    # Code is directly from vendor requests, using our modified request function here
    return request('delete', url, **kwargs)


_merge_docs(Session, vendor_requests_sessions.Session)
_merge_docs(session, vendor_requests_sessions.session)
_merge_docs(request, vendor_requests_api.request)
_merge_docs(get, vendor_requests_api.get)
_merge_docs(options, vendor_requests_api.options)
_merge_docs(head, vendor_requests_api.head)
_merge_docs(post, vendor_requests_api.post)
_merge_docs(put, vendor_requests_api.put)
_merge_docs(patch, vendor_requests_api.patch)
_merge_docs(delete, vendor_requests_api.delete)
